/*
 *   Copyright (c) 2005-2016 IBM Corporation.  All rights reserved.
 *
 *  Function:	- example: ping-pong benchmark
 *
 *	 Usage:		mpirun -f appfile -- [nbytes]
 *			(default is 256k)
 *
 *	 Notes:		- Define CHECK macro to check data integrity.
 *			runs ping-pong between several rank pairs
 *			simultaneously.  Only one rank on each host
 *			is used since this test is only interested in
 *			checking the interconnect.  Each rank is paired
 *			with self+1, then self+2, then self+4 .. then
 *			suspicious looking ranks are printed.
 *
 *  			"Suspicious" means any rank whose performance
 *			was half the global average, or who were in the
 *			bottom 10 ranks (but exempt if they were at
 *			least 80% of the global average perf).
 *
 *			And the above identification of bad ranks is done
 *			for two "performance" metrics.  One is the average
 *			over all the ranks this rank communicated with,
 *			and one is the minimum performance over all the
 *			ranks it communicated with.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <mpi.h>

#define NLOOPS		1000
#define ALIGN		4096

#define SEND(t)		MPI_Send(buf, nbytes, MPI_CHAR, partner, (t), \
				comm)
#define RECV(t)		MPI_Recv(buf, nbytes, MPI_CHAR, partner, (t), \
				comm, &status)
#ifdef CHECK
#  define SETBUF()	for (j=0; j<nbytes; j++) { \
				buf[j] = (char) (j + i); \
			}
#  define CLRBUF()	memset(buf, 0, nbytes)
#  define CHKBUF()	for (j = 0; j < nbytes; j++) { \
				if (buf[j] != (char) (j + i)) { \
					printf("error: buf[%d] = %d, " \
						"not %d\n", \
						j, buf[j], j + i); \
					break; \
				} \
			}
#else
#  define SETBUF()
#  define CLRBUF()
#  define CHKBUF()
#endif

double	**runtimes;
int	**peer;
int	ntimes = 0;

void
runiter(char *buf, int nbytes, MPI_Comm comm, int *sendto);

void pick_active_ranks(MPI_Comm *newcomm, char ***hostnames);
void send_to_offset(int *sendto, int nranks, int offset);
void print_bad_nodes(int nbytes, char **hostnames, MPI_Comm comm);

int
main(int argc, char *argv[])
{
	int		i;
	int		nbytes = 256 * 1024;
	int		myrank, nranks;
	char		*buf, *obuf;
	char		**hostnames;
	MPI_Comm	newcomm;
	int		*sendto;
	int		numrounds, offset;

	MPI_Init(&argc, &argv);

	pick_active_ranks(&newcomm, &hostnames);
	if (!newcomm) {
		MPI_Barrier(MPI_COMM_WORLD);
		MPI_Finalize();
		exit(0);
	}
	MPI_Comm_rank(newcomm, &myrank);
	MPI_Comm_size(newcomm, &nranks);

	if (nranks < 2) {
		if (myrank==0) printf("clustertest: Test REQUIRES two or more Hosts (not just ranks)\n");
		MPI_Barrier(MPI_COMM_WORLD);
		MPI_Finalize();
		exit(0);
	}

	if (myrank == 0) {
		numrounds = 1;
		for (offset=2; offset < nranks; /*offset*=2*/) {
			if (offset & (offset-1)) {
				offset = 4 * offset / 3;
			} else {
				offset = 3 * offset / 2;
			}
			++numrounds;
		}
		numrounds *= 2;
		runtimes = malloc(sizeof(double) * nranks * numrounds  +
			sizeof(double*) * numrounds);
		peer = malloc(sizeof(int) * nranks * numrounds  +
			sizeof(int*) * numrounds);
		if (!runtimes || !peer) {
			MPI_Abort(MPI_COMM_WORLD, MPI_ERR_BUFFER);
			exit(1);
		}
		for (i=0; i<numrounds; ++i) {
			runtimes[i] = (double*)(((char*)&runtimes[numrounds]) +
				i * nranks * sizeof(double));
			peer[i] = (int*)(((char*)&peer[numrounds]) +
				i * nranks * sizeof(int));
		}
	}

	sendto = malloc(nranks * sizeof(int));
	if (!sendto) {
		MPI_Abort(MPI_COMM_WORLD, MPI_ERR_BUFFER);
		exit(1);
	}

/*
 *  Take nbytes from command line if it's there.
 */
	for (i=1; i<argc; ++i) {
		nbytes = atoi(argv[i]);
	}
	if (nbytes < 0) nbytes = 256 * 1024;
/*
 * Page-align buffers and displace them in the cache to avoid collisions.
 */
	buf = (char *) malloc(nbytes + 524288 + (ALIGN - 1));
	obuf = buf;
	if (buf == 0) {
		MPI_Abort(MPI_COMM_WORLD, MPI_ERR_BUFFER);
		exit(1);
	}
	buf = (char *) ((((MPI_Aint) buf) + (ALIGN - 1)) & ~(ALIGN - 1));
	if (myrank > 0) buf += 524288;
	memset(buf, 0, nbytes);


	send_to_offset(sendto, nranks, 1);
	if (myrank==0) {
		printf("running ping pongs with offset 1\n");
	}
	runiter(buf, nbytes, newcomm, sendto);
	for (offset=2; offset < nranks; /*offset*=2*/) {
		send_to_offset(sendto, nranks, offset);
		if (myrank==0) {
			printf("running ping pongs with offset %d\n", offset);
		}
		runiter(buf, nbytes, newcomm, sendto);
		if (offset & (offset-1)) {
			offset = 4 * offset / 3;
		} else {
			offset = 3 * offset / 2;
		}
	}
	print_bad_nodes(nbytes, hostnames, newcomm);


	free(sendto);
	if (myrank == 0) {
		free(runtimes);
		free(peer);
		free(hostnames);
	}
	free(obuf);
	MPI_Comm_free(&newcomm);
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Finalize();
	exit(0);
}

void send_to_offset(int *sendto, int nranks, int offset) {
	int i;
	for (i=0; i<nranks; ++i) {
		sendto[i] = (i+offset)%nranks;
	}
}

void
pick_active_ranks(MPI_Comm *newcomm, char ***hostnames)
{
	int myrank, nranks; /* here means in comm world */
	int len, maxlen;
	int newmyrank, newnranks;
	int *whichranks;
	int i, j;
	char		myhost[MPI_MAX_PROCESSOR_NAME];

	char *partial_result_new = 0;
	char *partial_result_current;
	int partial_size_new = 2;
	int partner_offset = 1;

	MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
	MPI_Comm_size(MPI_COMM_WORLD, &nranks);
	MPI_Get_processor_name(myhost, &len);
	partial_result_current = myhost;

	int tmp = myrank;
	while (tmp % 2 == 0) { /* receiver */
		partial_result_new = malloc(partial_size_new *
			MPI_MAX_PROCESSOR_NAME);
		memcpy(partial_result_new, partial_result_current,
			partial_size_new/2*MPI_MAX_PROCESSOR_NAME);
		if (myrank + partner_offset < nranks) {
			MPI_Recv(partial_result_new +
				partial_size_new/2*MPI_MAX_PROCESSOR_NAME,
				partial_size_new/2*MPI_MAX_PROCESSOR_NAME,
				MPI_CHAR,
				myrank + partner_offset,
				99, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
		}
		partner_offset *= 2;
		partial_size_new *= 2;
		if (partial_result_current != myhost) {
			free(partial_result_current);
		}
		partial_result_current = partial_result_new;
		tmp /= 2;
		if (partner_offset >= nranks) {
			tmp = 1;
		}
	}
	/* else we become sender, one time */
	if (myrank != 0) {
		MPI_Send(partial_result_current,
			partial_size_new/2*MPI_MAX_PROCESSOR_NAME,
			MPI_CHAR, myrank - partner_offset, 99,
			MPI_COMM_WORLD);
		if (partial_result_current != myhost) {
			free(partial_result_current);
		}
	}

	newnranks = 0;
	whichranks = malloc(nranks * sizeof(int));
	if (myrank == 0) {
/*
 *  For each rank, look at it's hostname, and find the first rank
 *  in the list that uses that hostname.  If this is it, then put
 *  its id as the next rank we're using in whichranks[].
 */
		for (i=0; i<nranks; ++i) {
			for (j=0; j<=i; ++j) {
				if (!strcmp(
					partial_result_current +
						i*MPI_MAX_PROCESSOR_NAME,
					partial_result_current +
						j*MPI_MAX_PROCESSOR_NAME
				)) {
					if (j == i) {
						whichranks[newnranks++] = i;
					}
					j = nranks;
				}
			}
		}
		free(partial_result_current);
	}
	MPI_Bcast(&newnranks, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(whichranks, newnranks, MPI_INT, 0, MPI_COMM_WORLD);
	{
		MPI_Group g1, g2;
		MPI_Comm_group(MPI_COMM_WORLD, &g1);
        	MPI_Group_incl(g1, newnranks, whichranks, &g2);
        	MPI_Comm_create(MPI_COMM_WORLD, g2, newcomm);
        	MPI_Group_free(&g1);
        	MPI_Group_free(&g2);
	}
	free(whichranks);
	if (!(*newcomm)) { return; }
	MPI_Comm_rank(*newcomm, &newmyrank);
	MPI_Comm_size(*newcomm, &newnranks);

/*
 *  Consolidate all hostnames in an array at rank 0.
 *  The first nranks*sizeof(char*) bytes are an array of pointers
 *  into the string where the data lives.
 */
	MPI_Allreduce(&len, &maxlen, 1, MPI_INT, MPI_MAX, *newcomm);
	if (newmyrank == 0) {
		*hostnames = malloc((maxlen+1) * newnranks  +
			newnranks * sizeof(char*));
		if (*hostnames == 0) {
			MPI_Abort(MPI_COMM_WORLD, MPI_ERR_BUFFER);
			exit(1);
		}
	}
	MPI_Gather(myhost, maxlen+1, MPI_CHAR,
		(newmyrank==0)?(&(*hostnames)[newnranks]):0,
		maxlen+1, MPI_CHAR, 0, *newcomm);
	if (myrank==0) {
		for (i=0; i<newnranks; ++i) {
			(*hostnames)[i] =
				(char*)(&(*hostnames)[newnranks]) +
				i*(maxlen+1);
		}
	}
}

/*
 *  An incoming array will identify who each rank is sending to.
 *  Those sends will be done several at a time, allowing one rank
 *  from each block of X ranks to be a sender at the same time.
 */
void
runiter(char *buf, int nbytes, MPI_Comm comm, int *sendto)
{
	int i, j;
	int myrank, nranks;
	int blocksize;
	int *used;
#define PAST_SENDER 1
#define CURRENT_SENDER 2
#define CURRENT_RECVER 4
	MPI_Status	status;
	double start, stop;
	double mytime = -1;
	int allused;

	MPI_Comm_rank(comm, &myrank);
	MPI_Comm_size(comm, &nranks);

	used = malloc(nranks * sizeof(int));
	if (used == 0) {
		MPI_Abort(MPI_COMM_WORLD, MPI_ERR_BUFFER);
		exit(1);
	}
	for (i=0; i<nranks; ++i) {
		used[i] = 0;
	}

	blocksize = 1;
	while ((blocksize+1) * (blocksize+1) < nranks) {
		blocksize *= 2;
	}

	allused = 0;
	while (!allused) {
		int partner = -1;
/*
 *  For each round, each rank actually identifies all senders
 *  and receivers, and sees if it's one of them.
 */
		for (i=0; i<nranks; i+=blocksize) {
			j=0;
			while (j<blocksize && i+j<nranks &&
				(used[i+j] &
				  (PAST_SENDER|CURRENT_SENDER|CURRENT_RECVER)||
				(used[sendto[i+j]] & CURRENT_SENDER))
			) {
				++j;
			}
			if (j<blocksize && i+j<nranks) { /* found a pair */
				used[i+j] ^= CURRENT_SENDER;
				used[sendto[i+j]] ^= CURRENT_RECVER;
				if (myrank == i+j) {
					partner = sendto[i+j];
				}
				if (myrank == sendto[i+j]) {
					partner = i+j;
				}
			}
		}
		if (partner >= 0 && used[myrank]&CURRENT_SENDER) {
/*
 * warm-up loop
 */
			for (i = 0; i < 5; i++) {
				SEND(1);
				RECV(1);
			}
/*
 * timing loop
 */
			start = MPI_Wtime();
			for (i = 0; i < NLOOPS; i++) {
				SETBUF();
				SEND(1000 + i);
				CLRBUF();
				RECV(2000 + i);
				CHKBUF();
			}
			stop = MPI_Wtime();
			mytime = stop-start;
		}
		if (partner >= 0 && used[myrank]&CURRENT_RECVER) {
			for (i = 0; i < 5; i++) {
				RECV(1);
				SEND(1);
			}
			for (i = 0; i < NLOOPS; i++) {
				CLRBUF();
				RECV(1000 + i);
				CHKBUF();
				SETBUF();
				SEND(2000 + i);
			}
		}
		MPI_Barrier(comm);
		for (i=0; i<nranks; ++i) {
			if (used[i] & (CURRENT_SENDER|PAST_SENDER)) {
				used[i] = PAST_SENDER;
			} else {
				used[i] = 0;
			}
		}
		allused = 1;
		for (i=0; i<nranks; ++i) {
			if (!(used[i]&PAST_SENDER)) { allused = 0; }
		}
	}
	free(used);
	MPI_Gather(&mytime, 1, MPI_DOUBLE,
		(myrank==0)?(runtimes[ntimes]):0, 1, MPI_DOUBLE, 0, comm);
	if (myrank == 0) {
		for (i=0; i<nranks; ++i) {
			peer[ntimes][i] = sendto[i];
			runtimes[ntimes+1][sendto[i]] = runtimes[ntimes][i];
			peer[ntimes+1][sendto[i]] = i;
		}
	}
	ntimes += 2;
}

/*
 *  Sort high to low for qsort.
 */
double *scores;
int
dcomp(const void *a, const void *b)
{
	if ( scores[*((int*)a)] > scores[*((int*)b)] ) {
		return -1;
	} else if ( scores[*((int*)a)] < scores[*((int*)b)] ) {
		return 1;
	}
	return 0;
}

void
print_bad_nodes(int nbytes, char **hostnames, MPI_Comm comm)
{
	int myrank, nranks, i, j;
	double sum = 0, avg;
	char *warnabout;
	int *sortedranks;
	int nonefound;

	MPI_Comm_rank(comm, &myrank);
	MPI_Comm_size(comm, &nranks);

	if (myrank != 0) {
		return;
	}

	for (i=0; i<ntimes; ++i) {
		for (j=0; j<nranks; ++j) {
			sum += runtimes[i][j];
		}
	}
	avg = sum / ntimes / nranks;
/*
 *  Look for "low" performing ranks by sorting an array of ranks by
 *  their worst time, and consider bad ranks to be those who have
 *  a low score of either
 *    - below 50% of the global average
 *    - one of the bottom 10 low scores
 *  as long as it's not within 20% of the global average, in which case
 *  it's too good to flag as bad.  And same given an array of average
 *  scores.
 */
	warnabout = malloc(nranks * sizeof(char));
	scores = malloc(nranks * sizeof(double));
	sortedranks = malloc(nranks * sizeof(int));
	if (!warnabout || !scores || !sortedranks) {
		MPI_Abort(MPI_COMM_WORLD, MPI_ERR_BUFFER);
		exit(1);
	}
	for (i=0; i<nranks; ++i) {
		warnabout[i] = 0;
	}
/* avgs */
	for (i=0; i<nranks; ++i) {
		scores[i] = 0;
		sortedranks[i] = i;
		for (j=0; j<ntimes; ++j) {
			scores[i] += runtimes[j][i];
		}
		scores[i] /= ntimes;
	}
	qsort(sortedranks, nranks, sizeof(int), dcomp);
	i = 0;
	while ((i<10 || scores[sortedranks[i]] > avg*2) && i<nranks) {
		if (0.8 * scores[sortedranks[i]] >= avg) {
			warnabout[sortedranks[i]] = 1;
		}
		++i;
	}
/* mins (min performance I mean, so max times) */
	for (i=0; i<nranks; ++i) {
		scores[i] = 0;
		sortedranks[i] = i;
		for (j=0; j<ntimes; ++j) {
			if (runtimes[j][i] > scores[i]) {
				scores[i] = runtimes[j][i];
			}
		}
	}
	qsort(sortedranks, nranks, sizeof(int), dcomp);
	i = 0;
	while ((i<10 || scores[sortedranks[i]] > avg*2) && i<nranks) {
		if (0.8 * scores[sortedranks[i]] >= avg) {
			warnabout[sortedranks[i]] = 1;
		}
		++i;
	}
/* print from warnabout array */
	if (nbytes == 0) {
		printf("Global avg performance %.2f usec\n",
			avg/NLOOPS/2*1024*1024);
	} else {
		printf("Global avg performance %.2f usec (%.2f MB/sec)\n",
			avg/NLOOPS/2*1024*1024,
			nbytes / (1024. * 1024.) / (avg / NLOOPS / 2));
	}
	printf("------------------------------------------------\n");
	nonefound=1;
	for (i=0; i<nranks; ++i) {
		int r = sortedranks[i];
		if (warnabout[r]) {
			double ravg, rworst = 0;
			nonefound=0;
			sum = 0;
			for (j=0; j<ntimes; ++j) {
				sum += runtimes[j][r];
				if (runtimes[j][r] > rworst) {
					rworst = runtimes[j][r];
				}
			}
			ravg = sum / ntimes;
			if (nbytes == 0) {
				printf("[%s] worst %.2fu"
					", avg %.2f\n",
					hostnames[r],
					rworst/NLOOPS/2*1024*1024,
					ravg/NLOOPS/2*1024*1024
				);
			} else {
				printf("[%s] worst %.2fu "
					"(%.2fM)"
					", avg %.2f (%.2fM)\n",
					hostnames[r],
					rworst/NLOOPS/2*1024*1024,
					nbytes / (1024. * 1024.) /
						(rworst/NLOOPS/2),
					ravg/NLOOPS/2*1024*1024,
					nbytes / (1024. * 1024.) /
						(ravg/NLOOPS/2)
				);
			}
		}
	}
	if (nonefound) {
		printf("no suspicious performing ranks found\n");
	}

	free(warnabout);
	free(scores);
	free(sortedranks);
}
